module Lexer(
  Token (..),
  lexStr
  )where

import Data.Char

data Token = TokLPar
           | TokRPar
           | TokQuote
           | TokNumber Double
           | TokStr String
           | TokCharacter Char
           | TokSymbol String
           | TokErr String
  deriving (Eq, Show)

data StrState = StrStart
              | StrInside
              | StrEscape

identifyString :: String -> StrState -> (String, Maybe String)
identifyString ('"':xs) StrStart = identifyString xs StrInside
identifyString xs StrStart = (xs, Nothing) -- will never occur
identifyString "" _ = ("", Nothing) -- early end of file
identifyString ('\\':xs) StrInside = identifyString xs StrEscape
identifyString ('"':xs) StrInside = (xs, Just "")
identifyString ('n':xs) StrEscape = case identifyString xs StrInside
  of (ys, Just b) -> (ys, Just ('\n':b))
     y -> y
identifyString ('\\':xs) StrEscape = case identifyString xs StrInside
  of (ys, Just b) -> (ys, Just ('\\':b))
     y -> y
identifyString ('/':xs) StrEscape = case identifyString xs StrInside
  of (ys, Just b) -> (ys, Just ('/':b))
     y -> y
identifyString ('"':xs) StrEscape = case identifyString xs StrInside
  of (ys, Just b) -> (ys, Just ('"':b))
     y -> y
identifyString ('b':xs) StrEscape = case identifyString xs StrInside
  of (ys, Just b) -> (ys, Just ('\b':b))
     y -> y
identifyString ('f':xs) StrEscape = case identifyString xs StrInside
  of (ys, Just b) -> (ys, Just ('\f':b))
     y -> y
identifyString ('r':xs) StrEscape = case identifyString xs StrInside
  of (ys, Just b) -> (ys, Just ('\r':b))
     y -> y
identifyString ('t':xs) StrEscape = case identifyString xs StrInside
  of (ys, Just b) -> (ys, Just ('\t':b))
     y -> y
identifyString ('u':w:x:y:z:xs) StrEscape | isHexDigit w &&
                                            isHexDigit x &&
                                            isHexDigit y &&
                                            isHexDigit z =
                                              case identifyString xs StrInside
                                                   of (ys, Just b) -> (ys, Just (chr (read ('0':'x':w:x:y:z:"")):b))
                                                      a -> a
identifyString (x:xs) StrEscape = (x:xs, Nothing)
identifyString (x:xs) StrInside = case identifyString xs StrInside
  of (ys, Nothing) -> (ys, Nothing)
     (ys, Just b) -> (ys, Just (x:b))

data NumState = NumStart
              | NumNeg
              | NumIntegralComponent
              | NumFractionalComponent
              | NumExponentialComponent

identifyNum :: String -> NumState -> (String, Maybe String)
identifyNum "" NumStart = ("", Nothing)
identifyNum ('-':xs) NumStart = identifyNum xs NumNeg
identifyNum ('0':'.':x:xs) NumStart | isDigit x = case identifyNum (x:xs) NumFractionalComponent
                                                       of (ys, Just zs) -> (ys, Just ('0':'.':zs))
                                                          y -> y
                                    | otherwise = (x:xs, Nothing)
identifyNum (x:xs) NumStart | x == '0' = (x:xs, Nothing)
                            | isDigit x = identifyNum (x:xs) NumIntegralComponent
                            | otherwise = (x:xs, Nothing)
identifyNum "" NumNeg = ("", Nothing)
identifyNum ('0':'.':x:xs) NumNeg | isDigit x = case identifyNum (x:xs) NumFractionalComponent
                                                     of (ys, Just zs) -> (ys, Just ('-':'0':'.':zs))
                                                        y -> y
identifyNum (x:xs) NumNeg | x == '0' = (x:xs, Nothing)
                          | isDigit x = case identifyNum (x:xs) NumIntegralComponent
                                             of (ys, Just zs) -> (ys, Just ('-':zs))
                                                y -> y
                          | otherwise = (x:xs, Nothing)
identifyNum ('.':x:xs) NumIntegralComponent | isDigit x = case identifyNum (x:xs) NumFractionalComponent
                                                               of (ys, Just zs) -> (ys, Just ('.':zs))
                                                                  y -> y
                                            | otherwise = ('.':x:xs, Just "")
identifyNum (w:x:y:xs) NumIntegralComponent | (w == 'e' || w == 'E') &&
                                              (x == '+' || x == '-') &&
                                                isDigit y =
                                                  case identifyNum (y:xs) NumExponentialComponent
                                                       of (ys, Just zs) -> (ys, Just (w:x:zs))
                                                          z -> z
                                            | (w == 'e' || w == 'E') &&
                                                isDigit x =
                                                case identifyNum (x:y:xs) NumExponentialComponent
                                                     of (ys, Just zs) -> (ys, Just (w:zs))
                                                        z -> z
identifyNum (x:y:"") NumIntegralComponent | (x == 'e' || x == 'E') &&
                                            isDigit y = ("", Just (x:y:""))
                                          | isDigit x && isDigit y = ("", Just (x:y:""))
                                          | isDigit x = (y:"", Just (x:""))
                                          | otherwise = (x:y:"", Just "")
identifyNum (x:xs) NumIntegralComponent | isDigit x = case identifyNum xs NumIntegralComponent
                                                           of (ys, Just zs) -> (ys, Just (x:zs))
                                                              y -> y
                                        | otherwise = (x:xs, Just "")
identifyNum (w:x:y:xs) NumFractionalComponent | (w == 'e' || w == 'E') &&
                                                (x == '+' || x == '-') &&
                                                isDigit y =
                                                  case identifyNum (y:xs) NumExponentialComponent
                                                       of (ys, Just zs) -> (ys, Just (w:x:zs))
                                                          z -> z
                                              | (w == 'e' || w == 'E') &&
                                                isDigit x =
                                                case identifyNum (x:y:xs) NumExponentialComponent
                                                     of (ys, Just zs) -> (ys, Just (w:zs))
                                                        z -> z
identifyNum (x:y:"") NumFractionalComponent | (x == 'e' || x == 'E') &&
                                              isDigit y = ("", Just (x:y:""))
                                            | isDigit x && isDigit y = ("", Just (x:y:""))
                                            | isDigit x = (y:"", Just (x:""))
                                            | otherwise = (x:y:"", Just "")
identifyNum (x:xs) NumFractionalComponent | isDigit x = case identifyNum xs NumFractionalComponent
                                                             of (ys, Just zs) -> (ys, Just (x:zs))
                                                                y -> y
                                          | otherwise = (x:xs, Just "")
identifyNum (x:xs) NumExponentialComponent | isDigit x = case identifyNum xs NumExponentialComponent
                                                              of (ys, Just zs) -> (ys, Just (x:zs))
                                                                 y -> y
                                           | otherwise = (x:xs, Just "")
identifyNum [] _ = ("", Just "")

lexStr :: String -> [Token]
lexStr "" = []
lexStr ('(':xs) = TokLPar : lexStr xs
lexStr (')':xs) = TokRPar : lexStr xs
lexStr ('"':xs) = case identifyString ('"':xs) StrStart
  of (ys, Nothing) -> TokErr "End of file reached before matching \"" :lexStr ys
     (ys, Just b) -> TokStr b:lexStr ys
lexStr ('\'':"") = [TokErr "End of file reached before matching '"]
lexStr ('\'':'(':xs) = TokQuote : lexStr ('(':xs)
lexStr ('\'':x:'\'':xs) = TokCharacter x : lexStr xs
lexStr (' ':xs) = lexStr xs
lexStr ('\t':xs) = lexStr xs
lexStr ('\n':xs) = lexStr xs
lexStr ('\r':xs) = lexStr xs
lexStr (x:xs) | isDigit x || x == '-' = case identifyNum (x:xs) NumStart
                                             of (ys, Nothing) -> TokErr "Failed to parse number" : lexStr ys
                                                (ys, Just b) -> TokNumber (read b) : lexStr ys
lexStr _ = []
